// -*-C++-*-
// bladeloader.cpp
//
//  The WorldForge Project
//  Copyright (C) 2000  The WorldForge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about WorldForge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//

#include <stdio.h>
#include <sys/stat.h>
#include <parserInternals.h>

#include "bladeloader.h"

#ifdef WIN32
#define snprintf _snprintf
#endif

CoalSAXLoaderState CoalBladeLoader::m_state;

int CoalBladeLoader::m_depth = 0;
int CoalBladeLoader::m_fileindex = 0;
BladeXMLFunc CoalBladeLoader::m_functable[COAL_BLADE_SCOPE_END];
CoalDatabase *CoalBladeLoader::m_static_db = NULL;

// Array of text names for all the CoalBladeScope elements above.
// Absolutely must keep this array in sync with the CoalBladeScope
// enum.
const char *blade_elements[] =
{ "NONE", "media", "objects", "map", "graphic", "areafill", "texture",
  "algorithm", "seed", "contents", "region", "object", "color",
  "geometry", "vertices", "coord", "offset", "name", "desc", "file", "graphicref", 
  "layer", "shape", "fill", "red", "green", "blue", "x", "y", "z"
};
int blade_element_count = sizeof (blade_elements) / sizeof (char*);


CoalBladeLoader::CoalBladeLoader (CoalDatabase &db)
    : CoalMapLoader (db)
{
    m_depth = 0;
    m_xmlDoc = NULL;

    // This is totally a hack!  Since we don't want to force the
    // parent CoalMapLoader class to declare m_db as static, we keep
    // our own static copy in sync with m_db.  Sorry 'bout that.  (c:
    m_static_db = &db;

    InitCallbacks ();
    InitFuncTable ();
}

void CoalBladeLoader::InitCallbacks ()
{
    // Initialize all function pointers to NULL
    memset (&m_saxfuncs, 0, sizeof (xmlSAXHandler));

    // Parsing callbacks
    m_saxfuncs.startDocument = startDocument;
    m_saxfuncs.endDocument = endDocument;
    m_saxfuncs.startElement = startElement;
    m_saxfuncs.endElement = endElement;
    m_saxfuncs.characters = characters;

    // Formatting callbacks
    m_saxfuncs.warning = warning;
    m_saxfuncs.error = error;
    m_saxfuncs.fatalError = fatalError;
}

void CoalBladeLoader::InitFuncTable ()
{
    // Initialize all function pointers to default function
    for (int i = 0; i < COAL_BLADE_SCOPE_END; i++)
        m_functable[i] = &bxfDefault;

    // Hook up XML elements to the appropriate callback function.
    // This makes it easier to add new elements and will simplify
    // routing for each element.  As we add new handler functions, we
    // must wire them into the function table below...
    m_functable[COAL_BLADE_SCOPE_GRAPHIC] = &bxfGraphic;
    m_functable[COAL_BLADE_SCOPE_AREAFILL] = &bxfAreaFill;
    m_functable[COAL_BLADE_SCOPE_REGION] = &bxfRegion;
    m_functable[COAL_BLADE_SCOPE_COORD] = &bxfCoord;
    m_functable[COAL_BLADE_SCOPE_FILE] = &bxfFile;
    m_functable[COAL_BLADE_SCOPE_GEOMETRY] = &bxfGeometry;
    m_functable[COAL_BLADE_SCOPE_OBJECT] = &bxfObject;
    //m_functable[COAL_BLADE_SCOPE_] = &;
}

CoalBladeLoader::~CoalBladeLoader ()
{
    CleanupLoader ();
}

void CoalBladeLoader::SetDatabase (CoalDatabase *db)
{
    printf ("BLADE: STATIC SetDatabase()\n");
    if (db)
    {
        m_db = db;
        m_static_db = db;
    }
}

// The BLADE loader supports a list of mapfiles to load as a batch.
// Parse the mapfile string and invoke LoadSingleMap() for each
// single map we find.  (Could probably improve the algorithm
// below quite a bit...)
bool CoalBladeLoader::LoadMap (const char *mapfile)
{
    if (!mapfile)
        return false;

    printf ("BLADE: Loading batch map %s...\n", mapfile);
    ResetLoader ();

    // Possible separators between files
    const char *delim = ":;,";

    // If we only have one file, just load it.
    char *nextfile = strpbrk (mapfile, delim);
    if (!nextfile)
    {
        printf ("BLADE: LOADING SINGLE FILE: %s\n", mapfile);
        return LoadSingleMap (mapfile);
    }

    bool success = true;

    // Copy string because we will mangle it with '\0's
    char *copy = (char*)malloc (strlen (mapfile));
    strcpy (copy, mapfile);

    char *eos = copy + strlen (copy);  // End Of String
    char *thisfile = copy;

    //printf ("BLADE: copy = %p, eos = %p, strlen = %d\n", copy, eos, strlen(copy));
    do
    {
        nextfile = strpbrk (thisfile, delim);
        if (nextfile)
        {
            *nextfile = '\0';
            nextfile++;
        }
        else
            nextfile = eos;  // Kind of a hack

        printf ("BLADE: NEXT BATCH FILE: %s\n", thisfile);

        // Return false if any of the maps in the batch failed.
        success &= LoadSingleMap (thisfile);
        ResetLoader ();
        if (nextfile) thisfile = nextfile;
        //printf ("BLADE: thisfile = %p, nextfile = %p, eos = %p\n",
        //        thisfile, nextfile, eos);
    }
    while (nextfile < eos);

    free (copy);
    return success;
}

bool CoalBladeLoader::LoadSingleMap (const char *mapfile)
{
    printf ("BLADE: Loading map %s...\n", mapfile);

    struct stat statbuf;
    if (stat(mapfile, &statbuf) != 0)
    {
        printf ("BLADE: Warning, unable to find mapfile %s\n", mapfile);
        return false;
    }

    bool ret = 0;
    xmlParserCtxtPtr context;

    context = xmlCreateFileParserCtxt(mapfile);
    if (context == NULL)
    {
        return false;
    }
    context->sax = &m_saxfuncs;
    //context->userData = user_data;

    xmlParseDocument(context);

    if (context->wellFormed)
        ret = true;
    else
        ret = false;

    context->sax = NULL;
    xmlFreeParserCtxt(context);

    return ret;
}

void CoalBladeLoader::DepthPad ()
{
    for (int i = 0; i < m_depth; i++)
        printf ("   ");
}

bool CoalBladeLoader::LoadMap (const char *mapbuffer, int buffersize)
{
    printf ("BLADE ERROR: LoadMap into a buffer not implemented!!\n");
    return false;
}

bool CoalBladeLoader::CleanupLoader ()
{
    if (m_xmlDoc)
        xmlFreeDoc (m_xmlDoc);

    return true;
}

void CoalBladeLoader::ResetLoader ()
{
    m_depth = 0;
    m_fileindex = 0;
    m_state.ClearState ();
}

// Maybe move this generic function to CoalMapLoader...?
int CoalBladeLoader::Trim(string &str)
{
    if (str.size() == 0)
        return 0;

    int newstart = str.find_first_not_of (" \t\r\n");
    int newend = str.find_last_not_of (" \t\r\n");

    // Check for empty (padding-only) strings
    if (newend == -1 || newstart == -1)
    {
        str.erase ();
        return 0;
    }

    str = str.substr (newstart, newend + 1);

    return str.size ();  // New length
}

void CoalBladeLoader::ResolveGraphicsRef (string &graphicref,
                                          CoalGraphic **ret_graphic)
{
    if (!ret_graphic)
        return;

    CoalGraphic *graphic;
    int gfxcount = m_static_db->GetGraphicCount ();
    for (int i = 0; i < gfxcount; i++)
    {
        graphic = m_static_db->GetGraphic (i);
        if (graphic && (graphic->name == graphicref))
        {
            *ret_graphic = graphic;
            break;
        }
    }
}

//********************* SAX Callbacks *********************

void CoalBladeLoader::startDocument (void *ctx)
{
    printf ("BLADE: --------\nBLADE: Starting XML Document parsing\n");
    m_depth++;
}

void CoalBladeLoader::endDocument (void *ctx)
{
    printf ("BLADE: Ending XML Document parsing\nBLADE: --------\n");
    m_depth--;
}

void CoalBladeLoader::startElement (void *ctx, const xmlChar *name,
                            const xmlChar **attr)
{
    m_depth++;

    FireElementHandler (name, (void*)attr);
}

void CoalBladeLoader::endElement (void *ctx, const xmlChar *name)
{
    m_depth--;

    // Clean up our mess when we leave each element
    CoalBladeScope scope = GetScope (name);
    if (scope == COAL_BLADE_SCOPE_GRAPHIC)
        m_state.curGraphic = NULL;
    else if (scope == COAL_BLADE_SCOPE_AREAFILL)
        m_state.curFill = NULL;
    else if (scope == COAL_BLADE_SCOPE_REGION)
        m_state.curRegion = NULL;
    else if (scope == COAL_BLADE_SCOPE_OBJECT)
        m_state.curObject = NULL;
    else if (scope == COAL_BLADE_SCOPE_COORD && m_state.curRegion)
        m_state.curRegion->AddCoord (m_state.curCoord);
}

void CoalBladeLoader::characters (void *ctx, const xmlChar *ch, int len)
{
    switch ((int)m_state.ownerScope)
    {
    case COAL_BLADE_SCOPE_GRAPHIC:
        HandleGraphicsContents (ch, len);
        break;
    case COAL_BLADE_SCOPE_AREAFILL:
        HandleAreaFillContents (ch, len);
        break;
    case COAL_BLADE_SCOPE_REGION:
        HandleRegionContents (ch, len);
        break;
    case COAL_BLADE_SCOPE_OBJECT:
        HandleObjectContents (ch, len);
        break;
    }
}

void CoalBladeLoader::warning (void *ctx, const char *msg, ...)
{
    printf ("BLADE: ***WARNING*** %s\n", msg);
}

void CoalBladeLoader::error (void *ctx, const char *msg, ...)
{
    printf ("BLADE: ***ERROR*** %s\n", msg);
}

void CoalBladeLoader::fatalError (void *ctx, const char *msg, ...)
{
    printf ("BLADE: ***FATAL ERROR*** %s\n", msg);
}

void CoalBladeLoader::HandleGraphicsContents (const xmlChar *contents, int len)
{
    if (!m_state.curGraphic)
        return;

    string stuff ((const char*)contents, len);
    Trim (stuff);

    if (stuff.empty())
        return;

    switch ((int)m_state.curScope)
    {
    case COAL_BLADE_SCOPE_NAME:
        m_state.curGraphic->name += stuff;
        break;
    case COAL_BLADE_SCOPE_DESC:
        m_state.curGraphic->desc += stuff;
        break;
    case COAL_BLADE_SCOPE_FILE:
        if (m_fileindex == -1)
            m_state.curGraphic->filename += stuff;
        else
            m_state.curGraphic->altfiles[m_fileindex] += stuff;
        break;
    case COAL_BLADE_SCOPE_COLOR:
        // Don't need anything here right now...
        break;
    case COAL_BLADE_SCOPE_RED:
        m_state.curGraphic->bkgdcolor.red = atoi (stuff.c_str());
        break;
    case COAL_BLADE_SCOPE_GREEN:
        m_state.curGraphic->bkgdcolor.green = atoi (stuff.c_str());
        break;
    case COAL_BLADE_SCOPE_BLUE:
        m_state.curGraphic->bkgdcolor.blue = atoi (stuff.c_str());
        break;

        // We know that <coord> elements only occur inside of <offset>
        // elements when we're in a <graphic> scope.
    case COAL_BLADE_SCOPE_X:
        m_state.curGraphic->offset.SetX(atof (stuff.c_str()));
        break;
    case COAL_BLADE_SCOPE_Y:
        m_state.curGraphic->offset.SetY(atof (stuff.c_str()));
        break;
    case COAL_BLADE_SCOPE_Z:
        m_state.curGraphic->offset.SetZ(atoi (stuff.c_str()));
        break;
    }
}

void CoalBladeLoader::HandleAreaFillContents (const xmlChar *contents, int len)
{
    if (!m_state.curFill)
        return;
    
    string stuff ((const char*)contents, len);
    Trim (stuff);

    if (m_state.curScope == COAL_BLADE_SCOPE_GRAPHICREF)
    {
        // Need to search graphics list for a graphic with the same
        // name as what's in stuff.
        //m_state.curFill->name += stuff;

        ResolveGraphicsRef (stuff, (CoalGraphic**)&m_state.curFill->graphic);

    }
    else
        m_state.curFill->name += stuff;
}

void CoalBladeLoader::HandleRegionContents (const xmlChar *contents, int len)
{
    if (!m_state.curRegion)
        return;

    string stuff ((const char*)contents, len);
    if (Trim (stuff) == 0)
        return;

    switch ((int)m_state.curScope)
    {
    case COAL_BLADE_SCOPE_NAME:
        m_state.curRegion->name += stuff;
        break;
    case COAL_BLADE_SCOPE_LAYER:
        m_state.curRegion->SetLayer (atoi (stuff.c_str()));
        break;
    case COAL_BLADE_SCOPE_GEOMETRY:
        break;
    case COAL_BLADE_SCOPE_FILL:
        m_state.curRegion->SetFill (m_static_db->FindFill (stuff));
        break;
    case COAL_BLADE_SCOPE_X:
        m_state.curCoord.SetX(atof (stuff.c_str()));
        break;
    case COAL_BLADE_SCOPE_Y:
        m_state.curCoord.SetY(atof (stuff.c_str()));
        break;
    case COAL_BLADE_SCOPE_Z:
        m_state.curCoord.SetZ(atoi (stuff.c_str()));
        break;
    }
}

void CoalBladeLoader::HandleObjectContents (const xmlChar *contents, int len)
{
    if (!m_state.curObject)
        return;

    string stuff ((const char*)contents, len);
    if (Trim (stuff) == 0)
        return;

    switch ((int)m_state.curScope)
    {
    case COAL_BLADE_SCOPE_NAME:
        m_state.curObject->name += stuff;
        break;
    case COAL_BLADE_SCOPE_GRAPHICREF:
        ResolveGraphicsRef (stuff, &m_state.curObject->graphic);
        break;
    case COAL_BLADE_SCOPE_X:
        m_state.curObject->anchor.SetX(atof (stuff.c_str()));
        break;
    case COAL_BLADE_SCOPE_Y:
        m_state.curObject->anchor.SetY(atof (stuff.c_str()));
        break;
    case COAL_BLADE_SCOPE_Z:
        m_state.curObject->anchor.SetZ(atof (stuff.c_str()));
        break;
    }
}

//********************* Blade Function Table Callbacks *********************

// Convert a text element name into a CoalBladeScope value
CoalBladeScope CoalBladeLoader::GetScope (const xmlChar *name)
{
    CoalBladeScope scope = COAL_BLADE_SCOPE_NONE;

    for (int i = 0; i < blade_element_count; i++)
    {
        if (!strcmp ((const char*)name, blade_elements[i]))
        {
            scope = (CoalBladeScope)i;
            break;
        }
    }

    return scope;
}

// Fire off the handler function in the BladeXMLFunc function table
void CoalBladeLoader::FireElementHandler (const xmlChar *name, void *extradata)
{
    m_state.curScope = GetScope (name);
    m_functable[m_state.curScope] (name, extradata);
}

void CoalBladeLoader::bxfDefault (const xmlChar *name, void *extradata)
{
    // Uncomment these lines for better debugging...
    //DepthPad ();
    //printf ("BLADE: Hit default for element <%s>\n", (const char*)name);
}

void CoalBladeLoader::bxfGraphic (const xmlChar *name, void *extradata)
{
    m_state.curGraphic = new CoalPolyGraphic;
    m_static_db->AddGraphic (m_state.curGraphic);
    m_state.ownerScope = COAL_BLADE_SCOPE_GRAPHIC;
}

void CoalBladeLoader::bxfAreaFill (const xmlChar *name, void *extradata)
{
    m_state.curFill = new CoalFill;
    m_static_db->AddFill (m_state.curFill);
    m_state.ownerScope = COAL_BLADE_SCOPE_AREAFILL;
}

void CoalBladeLoader::bxfRegion (const xmlChar *name, void *extradata)
{
    m_state.curRegion = new CoalRegion;
    m_static_db->AddRegion (m_state.curRegion);
    m_state.ownerScope = COAL_BLADE_SCOPE_REGION;
}

void CoalBladeLoader::bxfObject (const xmlChar *name, void *extradata)
{
    m_state.curObject = new CoalObject;
    m_static_db->AddObject (m_state.curObject);
    m_state.ownerScope = COAL_BLADE_SCOPE_OBJECT;
}

void CoalBladeLoader::bxfFile (const xmlChar *name, void *extradata)
{
    char *orient[] = { "w", "n", "e", "s", "nw", "ne", "se", "sw" };

    m_fileindex = -1;

    // Parse out the element's attributes
    const char **attr = (const char**)extradata;
    if (attr)
    {
        int index = 0;
        while (attr[index])
        {
            // Look for the 'orient' attribute
            if (!strcmp (attr[index], "orient"))
            {
                // Assign an index into CoalPolyGraphic::altfiles[]
                // based on the value of the 'orient' attribute.
                for (int i = 0; i < 8; i++)
                    if (!strcmp (attr[index + 1], orient[i]))
                        m_fileindex = i;

                break;
            }

            // Could check for other attributes here...

            index += 2;
        }
    }
}

void CoalBladeLoader::bxfGeometry (const xmlChar *name, void *extradata)
{
    char *shape[] = { "circle", "triangle", "square", "rectangle", "polygon" };

    // Parse out the element's attributes
    const char **attr = (const char**)extradata;
    if (attr && m_state.curRegion)
    {
        int index = 0;
        while (attr[index])
        {
            // Look for the 'shape' attribute
            if (!strcmp (attr[index], "shape"))
            {
                for (int i = 0; i < 5; i++)
                    if (!strcmp (attr[index + 1], shape[i]))
                        m_state.curRegion->SetType ((CoalShapeType)i);

                break;
            }

            // Could check for other attributes here...

            index += 2;
        }
    }
}

void CoalBladeLoader::bxfCoord (const xmlChar *name, void *extradata)
{
    // Hit a new coord, so reset it to zeros
    m_state.curCoord.SetCoord (0.0, 0.0, 0.0);
}

// Write out the current database to an XML file
bool CoalBladeLoader::SaveMap (const char *mapfile, int save_flags,
                               CoalShape *partial_map)
{
    printf ("BLADE: Saving database to %s\n", mapfile);

    LoadXMLTree ((CoalSaveFlag)save_flags, partial_map);

    printf ("Attempting to save map to %s\n", mapfile);
    int ret = xmlSaveFile (mapfile, m_xmlDoc);
    printf ("Done saving map file\n");
    return (ret != -1);
}

// Convenience function to build an XML node for a CoalColor
xmlNodePtr CoalBladeLoader::BuildColorNode (CoalColor &color, xmlNodePtr parent)
{
    char colorstr[10];
    xmlNodePtr colorNode = xmlNewChild (parent, NULL, (xmlChar*)"color", NULL);

    snprintf (colorstr, sizeof(colorstr), "%d", color.red);
    xmlNewTextChild (colorNode, NULL, (xmlChar*)"red", (xmlChar*)colorstr);
    snprintf (colorstr, sizeof(colorstr), "%d", color.green);
    xmlNewTextChild (colorNode, NULL, (xmlChar*)"green", (xmlChar*)colorstr);
    snprintf (colorstr, sizeof(colorstr), "%d", color.blue);
    xmlNewTextChild (colorNode, NULL, (xmlChar*)"blue", (xmlChar*)colorstr);

    return colorNode;
}

// Convenience function to build an XML node for a CoalCoord
xmlNodePtr CoalBladeLoader::BuildCoordNode (CoalCoord &coord, xmlNodePtr parent)
{
    char numberstr[10];
    xmlNodePtr coordNode;
    coordNode = xmlNewChild (parent, NULL, (xmlChar*)"coord", NULL);

    // Add <x>, <y>, and <z> elements
    snprintf (numberstr, sizeof(numberstr), "%d", (int)coord.GetX ());
    xmlNewTextChild (coordNode, NULL, (xmlChar*)"x", (xmlChar*)numberstr);
    snprintf (numberstr, sizeof(numberstr), "%d", (int)coord.GetY ());
    xmlNewTextChild (coordNode, NULL, (xmlChar*)"y", (xmlChar*)numberstr);
    if (coord.GetZ())
    {
        snprintf (numberstr, sizeof(numberstr), "%d", (int)coord.GetZ ());
        xmlNewTextChild (coordNode, NULL, (xmlChar*)"z", (xmlChar*)numberstr);
    }

    return coordNode;
}

xmlNodePtr CoalBladeLoader::GetMediaNode ()
{
    int i;
    // Write <media> element
    xmlNodePtr mediaNode = xmlNewNode (NULL, (xmlChar*)"media");

    xmlNodePtr graphicNode;
    CoalGraphic *nextGraphic;
    int count = m_db->GetGraphicCount ();
    printf ("  Found %d graphics in database\n", count);

    for (i = 0; i < count; i++)
    {
        nextGraphic = m_db->GetGraphic (i);
        printf ("    %p\n", nextGraphic);
        if (!nextGraphic)
            continue;

        // Create <graphic> and <name> elements
        graphicNode = xmlNewChild (mediaNode, NULL, (xmlChar*)"graphic", NULL);
        xmlSetProp (graphicNode, (xmlChar*)"type", (xmlChar*)"2d");
        xmlNewTextChild (graphicNode, NULL, (xmlChar*)"name", 
                                       (xmlChar*)nextGraphic->name.c_str());

        // Add <file> elements
        xmlNodePtr fileNode;
        if (!nextGraphic->filename.empty())
        {
            fileNode = xmlNewTextChild (graphicNode, NULL, (xmlChar*)"file", 
                                        (xmlChar*)nextGraphic->filename.c_str());
            xmlSetProp (fileNode, (xmlChar*)"orient", (xmlChar*)"base");
        }

        // Add optional <file> elements
        char *orient[] = { "w", "n", "e", "s", "nw", "ne", "se", "sw" };
        CoalPolyGraphic *poly = (CoalPolyGraphic*)nextGraphic;
        for (int j = 0; j < 8; j++)
        {
            if (!poly->altfiles[j].empty())
            {
                fileNode = xmlNewTextChild (graphicNode, NULL, (xmlChar*)"file", 
                                            (xmlChar*)poly->altfiles[j].c_str());
                xmlSetProp (fileNode, (xmlChar*)"orient", (xmlChar*)orient[j]);

            }
        }

        // Add <color> element
        BuildColorNode (nextGraphic->bkgdcolor, graphicNode);

        // Add optional <offset> element
        if (nextGraphic->offset != CoalCoord (0, 0, 0))
        {
            xmlNodePtr offsetNode;
            offsetNode = xmlNewChild (graphicNode, NULL, (xmlChar*)"offset", NULL);
            BuildCoordNode (nextGraphic->offset, offsetNode);
        }

        // Add <desc> element
        xmlNewTextChild (graphicNode, NULL, (xmlChar*)"desc",
                         (xmlChar*)nextGraphic->desc.c_str());
    }

    // Add <areafill> elements
    xmlNodePtr fillNode;
    xmlNodePtr textureNode;
    CoalFill *nextFill;
    count = m_db->GetFillCount ();
    printf ("  Found %d fills in database\n", count);

    for (i = 0; i < count; i++)
    {
        nextFill = m_db->GetFill (i);
        printf ("    %p\n", nextFill);
        if (!nextFill)
            continue;

        // Create <areafill> element and <name> subelement
        fillNode = xmlNewChild (mediaNode, NULL, (xmlChar*)"areafill", NULL);
        xmlNewTextChild (fillNode, NULL, (xmlChar*)"name", 
                         (xmlChar*)nextFill->name.c_str());

        if (nextFill->graphic)
        {
            textureNode = xmlNewChild (fillNode, NULL, (xmlChar*)"texture", NULL);
            xmlNewTextChild (textureNode, NULL, (xmlChar*)"graphicref",
                             (xmlChar*)nextFill->graphic->name.c_str());
        }
    }

    printf ("Done parsing <media> element\n");
    return mediaNode;
}

xmlNodePtr CoalBladeLoader::GetMapNode ()
{
    // Write <map> element
    xmlNodePtr mapNode = xmlNewNode (NULL, (xmlChar*)"map");

    // Add <region> elements
    xmlNodePtr regionNode;
    xmlNodePtr geometryNode;
    xmlNodePtr verticesNode;
    CoalRegion *nextRegion;
    int count = m_db->GetRegionCount ();
    printf ("  Found %d regions in database\n", count);

    char numberstr[10];
    for (int i = 0; i < count; i++)
    {
        nextRegion = m_db->GetRegion (i);
        printf ("    %p\n", nextRegion);
        if (!nextRegion)
            continue;

        // Create <region> element and <name> subelement
        regionNode = xmlNewChild (mapNode, NULL, (xmlChar*)"region", NULL);
        xmlNewTextChild (regionNode, NULL, (xmlChar*)"name", 
                         (xmlChar*)nextRegion->name.c_str());
        snprintf (numberstr, sizeof(numberstr), "%d", nextRegion->GetLayer ());
        xmlNewTextChild (regionNode, NULL, (xmlChar*)"layer", (xmlChar*)numberstr);

        char *shape[] = { "circle", "triangle", "square",
                          "rectangle", "polygon" };

        // Add <geometry> element and <vertices> subelement
        geometryNode = xmlNewChild (regionNode, NULL, (xmlChar*)"geometry", NULL);
        xmlSetProp (geometryNode, (xmlChar*)"shape",
                    (xmlChar*)shape[nextRegion->GetType()]);
        verticesNode = xmlNewChild (geometryNode, NULL, (xmlChar*)"vertices", NULL);

        // Add list of <coord> elements
        CoalCoord coord;
        int coordCount = nextRegion->GetCoordCount ();
        for (int j = 0; j < coordCount; j++)
        {
            coord = nextRegion->GetCoord (j);
            BuildCoordNode (coord, verticesNode);
        }

        // Add <fill> element
        CoalFill *fill = nextRegion->GetFill ();
        if (fill)
        {
            xmlNodePtr fillNode;
            fillNode = xmlNewTextChild (regionNode, NULL, (xmlChar*)"fill",
                                    (xmlChar*)fill->name.c_str());
        }
    }

    printf ("Done parsing <map> element\n");
    return mapNode;
}

xmlNodePtr CoalBladeLoader::GetObjectsNode ()
{
    printf ("before---\n");
    if (m_db->GetObjectCount() == 0)
        return NULL;
    printf ("after---\n");

    // Write <objects> element
    xmlNodePtr objectsNode = xmlNewNode (NULL, (xmlChar*)"objects");

    // Add <object> elements
    xmlNodePtr objectNode;
    CoalObject *nextObject;
    int count = m_db->GetObjectCount ();
    printf ("  Found %d objects in database\n", count);

    for (int i = 0; i < count; i++)
    {
        nextObject = m_db->GetObject (i);
        printf ("    %p\n", nextObject);
        if (!nextObject)
            continue;

        // Create <object> element and <name> subelement
        objectNode = xmlNewChild (objectsNode, NULL, (xmlChar*)"object", NULL);
        xmlNewTextChild (objectNode, NULL, (xmlChar*)"name", 
                         (xmlChar*)nextObject->name.c_str());

        BuildCoordNode (nextObject->anchor, objectNode);

        if (nextObject->graphic)
        {
            xmlNewTextChild (objectNode, NULL, (xmlChar*)"graphicref",
                             (xmlChar*)nextObject->graphic->name.c_str());
        }
    }

    printf ("Done parsing <objects> element\n");
    return objectsNode;
}

// Sync up m_xmlDoc to current contents of database.  Mainly used
// for SaveMap().
bool CoalBladeLoader::LoadXMLTree (CoalSaveFlag flags, CoalShape *partial_map)
{
    printf ("BLADE: Refreshing XML Tree\n");

    if (m_xmlDoc)
        xmlFreeDoc (m_xmlDoc);

    if (partial_map)
        printf ("WARNING: Ignoring request to save partial map -- "
                "not yet implemented!!");

    // Create new XML document and top-level <blade> element
    m_xmlDoc = xmlNewDoc ((xmlChar*)"1.0");
    xmlNodePtr rootNode = xmlNewNode (NULL, (xmlChar*)"blade");
    xmlDocSetRootElement (m_xmlDoc, rootNode);

    // Add <!DOCTYPE blade SYSTEM "blade.dtd"> instruction
    // FIXME: This doesn't seem to work...
    //xmlNewDtd (m_xmlDoc, (xmlChar*)"blade", (xmlChar*)NULL,
    //           (xmlChar*)"blade.dtd");

    printf ("Loading XML tree with flags %d\n", flags);

    if (flags & COAL_SAVE_MEDIA)
        xmlAddChild (rootNode, GetMediaNode ());

    if (flags & COAL_SAVE_MAP)
        xmlAddChild (rootNode, GetMapNode ());

    if (flags & COAL_SAVE_OBJECTS)
        xmlAddChild (rootNode, GetObjectsNode ());

    printf ("Done refreshing XML tree\n");
    return true;
}
