// -*-C++-*-
// bladeloader.h
//
//  The WorldForge Project
//  Copyright (C) 2000  The WorldForge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about WorldForge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//

// Load BLADE XML files.  This is a pretty messy class, internally,
// but all this should be transparent to the user, aside from the
// warning below about using multiple CoalBladeLoader objects at the
// same time.

// On one hand, it's kind of silly to create a C++ class like this and
// set most of its contents to be static members.  However, in the
// case of SAX parsing with the gnome-xml, we're pretty much stuck
// with it if we want to include the callback functions inside the
// class.  This in turn means that all member variables any callback
// function references must also be static, which leads to nearly
// everything being static.
//
// Yet, it's still worthwhile to do this as a C++ class, since we are
// making use of inheritance and virtual functions.  We must still
// implement the LoadMap() and other essential CoalMapLoader
// functions.  However, we should consider the CoalBladeLoader to be a
// singleton class (i.e. only one can exist at a time).  Consequently,
// we'll have to be careful to Reset the function table each time we
// call LoadMap().  Fortunately, this class only implements one
// significant API call, so we don't have to worry as much about
// interleaving API calls from different CoalBladeLoader objects.

#ifndef __BLADELOADER_H
#define __BLADELOADER_H

#include <tree.h>
#include <parser.h>
#include "maploader.h"

typedef enum
{
    COAL_BLADE_SCOPE_NONE,

    // High-level elements
    COAL_BLADE_SCOPE_MEDIA,
    COAL_BLADE_SCOPE_OBJECTS,
    COAL_BLADE_SCOPE_MAP,

    // <media> sub-elements
    COAL_BLADE_SCOPE_GRAPHIC,
    COAL_BLADE_SCOPE_AREAFILL,
    COAL_BLADE_SCOPE_TEXTURE,
    COAL_BLADE_SCOPE_ALGORITHM,
    COAL_BLADE_SCOPE_SEED,
    COAL_BLADE_SCOPE_CONTENTS,

    // <objects> sub-elements
    // FIXME: none yet...

    // <map> sub-elements
    COAL_BLADE_SCOPE_REGION,
    COAL_BLADE_SCOPE_OBJECT,

    // Generic elements
    COAL_BLADE_SCOPE_COLOR,
    COAL_BLADE_SCOPE_GEOMETRY,
    COAL_BLADE_SCOPE_VERTICES,
    COAL_BLADE_SCOPE_COORD,
    COAL_BLADE_SCOPE_OFFSET,

    // Leaf nodes (have text contents)
    COAL_BLADE_SCOPE_NAME,
    COAL_BLADE_SCOPE_DESC,
    COAL_BLADE_SCOPE_FILE,
    COAL_BLADE_SCOPE_GRAPHICREF,
    COAL_BLADE_SCOPE_LAYER,
    COAL_BLADE_SCOPE_SHAPE,
    COAL_BLADE_SCOPE_FILL,
    COAL_BLADE_SCOPE_RED,
    COAL_BLADE_SCOPE_GREEN,
    COAL_BLADE_SCOPE_BLUE,
    COAL_BLADE_SCOPE_X,
    COAL_BLADE_SCOPE_Y,
    COAL_BLADE_SCOPE_Z,

    COAL_BLADE_SCOPE_END  // Total # of enums
} CoalBladeScope;

// Array of text names for all the CoalBladeScope elements above.
// Absolutely must keep this array in sync with the CoalBladeScope
// enum.
extern const char *blade_elements[];

// Prototype for Blade XML function table
typedef void (*BladeXMLFunc) (const xmlChar *name, void *extradata);

// Modifiers to the way we save the map.  We don't define how much of
// the map to save here, since we supply that in the partial_map
// parameter to SaveMap().  A category is a 2nd-level blade element, e.g.,
// <media>, <map>, <objects>.
typedef enum
{
    COAL_SAVE_MEDIA =       1 << 0,  // Save <media> element
    COAL_SAVE_MAP =         1 << 1,  // Save <media> element
    COAL_SAVE_OBJECTS =     1 << 2,  // Save <media> element

    // Save all categories
    COAL_SAVE_ALL = COAL_SAVE_MEDIA | COAL_SAVE_MAP | COAL_SAVE_OBJECTS
} CoalSaveFlag;

// Static state of SAX loader (referenced from static
// callback functions).  Could probably move this to
// CoalMapLoader.
class CoalSAXLoaderState
{
public:
    CoalSAXLoaderState () { ClearState (); }
    void ClearState ()
        {
            curScope = ownerScope = COAL_BLADE_SCOPE_NONE;
            curGraphic = NULL;
            curFill = NULL;
            curRegion = NULL;
            curObject = NULL;
        }

    CoalBladeScope curScope;
    CoalBladeScope ownerScope;

    CoalPolyGraphic *curGraphic;
    CoalFill *curFill;
    CoalRegion *curRegion;
    CoalObject *curObject;
    CoalCoord curCoord;
};

class CoalBladeLoader : public CoalMapLoader
{
public:
    CoalBladeLoader (CoalDatabase &db);
    virtual ~CoalBladeLoader ();

    virtual bool LoadMap (const char *mapfile);
    virtual bool LoadMap (const char *mapbuffer, int buffersize);
    virtual bool SaveMap (const char *mapfile, int save_flags = COAL_SAVE_ALL,
                          CoalShape *partial_map = NULL);
    virtual bool CleanupLoader ();
    virtual void SetDatabase (CoalDatabase *db);

private:
    static int Trim(string &str);
    static void DepthPad ();
    void InitCallbacks ();
    void InitFuncTable ();
    static CoalBladeScope GetScope (const xmlChar *name);
    static void FireElementHandler (const xmlChar *name, void *extradata);
    static void ResolveGraphicsRef (string &graphicref,
                                    CoalGraphic **ret_graphic);
    bool LoadSingleMap (const char *mapfile);
    void ResetLoader ();

    // XML Maintenance
    bool LoadXMLTree (CoalSaveFlag flags, CoalShape *partial_map = NULL);
    xmlNodePtr BuildColorNode (CoalColor &color, xmlNodePtr parent);
    xmlNodePtr BuildCoordNode (CoalCoord &coord, xmlNodePtr parent);
    xmlNodePtr GetMediaNode ();
    xmlNodePtr GetMapNode ();
    xmlNodePtr GetObjectsNode ();

private:
    // SAX Callbacks
    static void startDocument (void *ctx);
    static void endDocument (void *ctx);
    static void startElement (void *ctx, const xmlChar *name,
                              const xmlChar **attr);
    static void endElement (void *ctx, const xmlChar *name);
    static void characters (void *ctx, const xmlChar *ch,
		            int len);
    static void warning (void *ctx, const char *msg, ...);
    static void error (void *ctx, const char *msg, ...);
    static void fatalError (void *ctx, const char *msg, ...);

    // SAX helper functions
    static void HandleGraphicsContents (const xmlChar *contents, int len);
    static void HandleAreaFillContents (const xmlChar *contents, int len);
    static void HandleRegionContents (const xmlChar *contents, int len);
    static void HandleObjectContents (const xmlChar *contents, int len);

    // Blade XML Function Table Callbacks (add functions here
    // to match new tags or new parsing behavior.
    static void bxfDefault (const xmlChar *name, void *extradata);

    // Elements with other elements as contents (usually need to
    // allocate a new instance here).
    static void bxfGraphic (const xmlChar *name, void *extradata);
    static void bxfAreaFill (const xmlChar *name, void *extradata);
    static void bxfRegion (const xmlChar *name, void *extradata);
    static void bxfObject (const xmlChar *name, void *extradata);

    // Elements with attributes that need parsing
    static void bxfFile (const xmlChar *name, void *extradata);
    static void bxfGeometry (const xmlChar *name, void *extradata);

    // Elements that need to be initialized when first hit
    static void bxfCoord (const xmlChar *name, void *extradata);

private:
    xmlSAXHandler m_saxfuncs;
    xmlDocPtr m_xmlDoc;

    static BladeXMLFunc m_functable[COAL_BLADE_SCOPE_END];
    static int m_depth;
    static int m_fileindex;

    static CoalDatabase *m_static_db;
    static CoalSAXLoaderState m_state;
};

#endif  // __BLADELOADER_H
