// -*-C++-*-
// database.cpp
//
//  The WorldForge Project
//  Copyright (C) 2000  The WorldForge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about WorldForge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//

#include <math.h>
#include "database.h"
#include "debug.h"

CoalDatabase::CoalDatabase ()
{
    m_verbose = false;
    SetTileSize (1.0, 1.0);
}

CoalDatabase::~CoalDatabase ()
{
}

// Get the exact shape of all regions in the map (i.e. a silhouette).
// FIXME: Doesn't quite work yet...
void CoalDatabase::GetMapShape (CoalShape &shape)
{
    shape.ClearShape ();

    CoalRegionArray regions;
    GetSortedRegions (regions);

    if (m_verbose)
        printf ("Scanning %d regions to calculate map shape\n", regions.size ());

    unsigned int count = regions.size ();
    if (count == 0)
        return;

    shape = *(CoalShape*)(regions[0]);
    for (unsigned int i = 1; i < count; i++)
    {
        // FIXME: Should get intersection of all the regions in the
        // CoalRegionArray, not just the last one.  This will require
        // a new CoalShape::Intersect() function.
        // 
        //shape.Intersect (*(CoalShape*)(regions[i]));

        shape = *(CoalShape*)(regions[i]);
        //debug.Dump (shape);
    }
}

// Get a bounding rect that encloses all regions in the map.
void CoalDatabase::GetMapExtents (CoalRect &rect)
{
    rect.ClearRect ();

    CoalRegionArray regions;
    GetSortedRegions (regions);

    if (m_verbose)
        printf ("Scanning %d regions to calculate map shape\n", regions.size ());

    unsigned int count = regions.size ();
    if (count == 0)
        return;

    // Set rect to the bounds of the first region.
    regions[0]->GetBoundingRect (rect);
    if (m_verbose)
    {
        printf ("Initial rect: \n\t");
        CoalDebug::Dump(rect);
    }

    // Add the bounding regions of the remaining regions to the first.
    CoalRect newrect;
    for (unsigned int i = 1; i < count; i++)
    {
        regions[i]->GetBoundingRect (newrect);
        rect.Intersect (newrect);

        if (m_verbose)
        {
            printf ("Rect %d: \n\t", i);
            CoalDebug::Dump(newrect);
        }
    }

    if (m_verbose)
    {
        printf ("Dumping map extents...\n\t");
        CoalDebug::Dump(rect);
    }
}

void CoalDatabase::GetMapPixelSize (int &width, int &height)
{
    CoalCoordType twidth, theight;
    GetTileSize (twidth, theight);

    // Get rectangle bounding area
    CoalRect rect;
    GetMapExtents (rect);

    width = (int)(rect.Width () * twidth);
    height = (int)(rect.Height () * theight);

    if (m_verbose)
    {
        printf ("tiles (%f, %f)\nBounding rect:\n", twidth, theight);
        CoalDebug::Dump (rect);
        printf ("map size = %d, %d\n", width, height);
    }
}

void CoalDatabase::GetMapOrigin (CoalCoord &origin)
{
    CoalRect rect;
    GetMapExtents (rect);
    origin = rect.GetTopLeft ();
}

// Get a single tile
CoalTile* CoalDatabase::GetTileAt (const CoalCoord &coord)
{
    // Get list of regions to search for a point hit.
    CoalRegionArray regions;
    GetSortedRegions (regions);

    for (unsigned int i = 0; i < regions.size (); i++)
    {
        if (regions[i] && regions[i]->IsPointIn (coord))
        {
            if (0 && m_verbose) printf ("Found hit in region %d\n", i);

            CoalFill *fill = regions[i]->GetFill ();
            if (fill)
            {
                if (0 && m_verbose)
                {
                    printf ("  Creating tile");
                    if (fill->graphic)
                        printf (", %s", fill->graphic->filename.c_str());
                    printf ("\n");
                }
                CoalTile *tile = new CoalTile;
                tile->graphic = fill->graphic;
                tile->coord = coord;
                return tile;
            }
        }
    }
    return NULL;
}

// FIXME: Should really follow the _real_ CoalShape, not the bounding
// rect of that shape.  (This is a quick fix.)
void CoalDatabase::GetTilesIn (const CoalShape &shape, CoalTileArray &tiles)
{
    CoalRect rect;
    shape.GetBoundingRect (rect);

    GetTilesIn (rect, tiles);
}

// Get a list of all tiles in a polygonal region.  The tiles array
// will be populated with pointers to newly-created CoalTile's.  It is
// up to the calling application to free up the contents of the
// CoalTileArray.
void CoalDatabase::GetTilesIn (const CoalRect &rect, CoalTileArray &tiles)
{
    tiles.clear ();

    // Make horizontal scanline sweeps across the inside of the shape.

    if (m_verbose)
        printf ("GetTilesIn: before (%f, %f, %f, %f) ",
                rect.GetLeft (), rect.GetTop (),
                rect.GetRight (), rect.GetBottom ());

    CoalRect xrect = rect;
#if 0
    // Expand rect so that it encompasses all of the tile squares that
    // it intersects.
    xrect.SetTop (floor (xrect.GetTop () / m_tileheight) * m_tileheight);
    xrect.SetLeft (floor (xrect.GetLeft () / m_tilewidth)  * m_tilewidth);
    xrect.SetBottom (ceil (xrect.GetBottom () / m_tileheight) * m_tileheight);
    xrect.SetRight (ceil (xrect.GetRight () / m_tilewidth)  * m_tilewidth);
#endif

    if (m_verbose)
        printf ("after (%f, %f, %f, %f)\n",
                xrect.GetLeft (), xrect.GetTop (),
                xrect.GetRight (), xrect.GetBottom ());

    CoalTile *tile;
    CoalCoordType x;
    CoalCoordType y;
    for (y = xrect.GetTop (); y < xrect.GetBottom (); y++ /* += m_tileheight*/)
    {
        for (x = xrect.GetLeft (); x < xrect.GetRight (); x++ /* += m_tilewidth*/)
        {
            tile = GetTileAt (CoalCoord (x, y));
            if (tile)
                tiles.push_back (tile);
            /* printf ("  Got tile at (%f, %f)\n", x, y); */
        }
    }
    if (m_verbose) printf ("Got %d tiles...\n", tiles.size ());
}

// Get a list of tiles that match a customizable query
// (Should perform a Traversal (q.v. Design Patterns)
void CoalDatabase::QueryTiles (const CoalQuery &query, CoalTileArray &tiles)
{
    // FIXME: Not yet implemented
}

void CoalDatabase::FreeTiles (CoalTileArray &tiles)
{
    unsigned int tilecount = tiles.size ();

    // Don't delete the CoalGraphic pointer!  That's still in use.
    for (unsigned int i = 0; i < tilecount; i++)
        delete tiles[i];

    tiles.clear ();
}

// Get an array of all regions, sorted by layer.
void CoalDatabase::GetSortedRegions (CoalRegionArray &regions)
{
    regions.clear ();

    // FIXME: Not really sorted by layer!!  For now, just return in
    // the reverse order they appeared in the map (i.e. the later
    // regions will get hit first when iterating through the
    // CoalRegionArray).
    for (int i = GetRegionCount () - 1; i >= 0; i--)
    {
        regions.push_back (GetRegion (i));
    }
}

void CoalDatabase::GetAllTiles (CoalTileArray &tiles)
{
    CoalRect rect;

    GetMapExtents (rect);
    GetTilesIn (rect, tiles);
}

// Search sequentially through object array and find the first object
// at exactly those coords.
//
// FIXME: Should test for hits within the boundary of the object, not
// just the anchor point.
CoalObject* CoalDatabase::GetObjectAt (const CoalCoord &coord)
{
    int count = GetObjectCount ();
    for (int i = 0; i < count; i++)
    {
        if (m_objects[i] && (m_objects[i]->anchor == coord))
            return m_objects[i];
    }

    return NULL;
}

// Cycle through entire list of objects, and add each object whose
// anchor point falls within 'shape' into the 'objects' array.
void CoalDatabase::GetObjectsIn (const CoalShape &shape, CoalObjectArray &objects)
{
    objects.clear ();

    int count = GetObjectCount ();
    for (int i = 0; i < count; i++)
    {
        if (m_objects[i] && (shape.IsPointIn (m_objects[i]->anchor)))
            objects.push_back (m_objects[i]);
    }
}

// ********** Container manipulation **********

int CoalDatabase::GetGraphicCount ()
{
    return m_graphics.size ();
}

bool CoalDatabase::AddGraphic (CoalGraphic *graphic)
{
    if (!graphic)
	return false;

    m_graphics.push_back (graphic);
    graphic->id = GetGraphicCount ();

    return true;
}

#if 0
bool CoalDatabase::RemoveGraphic (GID gid)
{
    CoalGraphicMapIter iter = m_graphics.find(gid);
    if (iter == m_graphics.end())
	return false;

    m_graphics.erase (iter);
    return true;
}
#endif

CoalGraphic* CoalDatabase::GetGraphic (unsigned int index)
{
    if (index >= (unsigned int)GetGraphicCount ())
        return NULL;

    return m_graphics[index];
}

int CoalDatabase::GetFillCount ()
{
    return m_fills.size ();
}

bool CoalDatabase::AddFill (CoalFill *fill)
{
    if (!fill)
	return false;

    m_fills.push_back (fill);
    return true;
}

#if 0
bool CoalDatabase::RemoveFill (CoalFill *fill)
{
    FillArray::iterator iter = m_fills.find(fill);
    if (iter == m_fills.end())
	return false;

    m_fills.erase (iter);
    return true;
}
#endif

CoalFill* CoalDatabase::GetFill (unsigned int index)
{
    if (index >= (unsigned int)GetFillCount ())
        return NULL;

    return m_fills[index];
}

CoalFill* CoalDatabase::FindFill (string &fillname)
{
    int count = GetFillCount ();
    for (int i = 0; i < count; i++)
    {
        if (m_fills[i] && m_fills[i]->name == fillname)
            return m_fills[i];
    }

    return NULL;
}

int CoalDatabase::GetRegionCount ()
{
    return m_regions.size ();
}

bool CoalDatabase::AddRegion (CoalRegion *region)
{
    if (!region)
	return false;

    m_regions.push_back (region);
    return true;
}

#if 0
bool CoalDatabase::RemoveRegion (CoalRegion *region)
{
    RegionArray::iterator iter = m_regions.find(region);
    if (iter == m_regions.end())
	return false;

    m_regions.erase (iter);
    return true;
}
#endif

CoalRegion* CoalDatabase::GetRegion (unsigned int index)
{
    if (index >= (unsigned int)GetRegionCount ())
        return NULL;

    return m_regions[index];
}

int CoalDatabase::GetObjectCount ()
{
    return m_objects.size ();
}

bool CoalDatabase::AddObject (CoalObject *object)
{
    if (!object)
	return false;

    m_objects.push_back (object);
    return true;
}

#if 0
bool CoalDatabase::RemoveObject (CoalObject *object)
{
    ObjectArray::iterator iter = m_objects.find(object);
    if (iter == m_objects.end())
	return false;

    m_objects.erase (iter);
    return true;
}
#endif

CoalObject* CoalDatabase::GetObject (unsigned int index)
{
    if (index >= (unsigned int)GetObjectCount ())
        return NULL;

    return m_objects[index];
}

