// -*-C++-*-
// shape.cpp
//
//  The WorldForge Project
//  Copyright (C) 2000  The WorldForge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about WorldForge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//

#include <stdio.h>  // TEMP: just here for FIXME messages

#include "shape.h"
#include "debug.h"

CoalShape::~CoalShape ()
{
}

void CoalShape::ClearShape ()
{
    m_coords.clear ();
    m_length = 0.0;
    m_type = COAL_SHAPE_RECT;
}

CoalCoord CoalShape::GetCoord (unsigned int index) const
{
    if (index < 0 || index >= m_coords.size ())
    {
        CoalCoord badCoord;
        badCoord.SetInvalid ();
        return badCoord;
    }

    return m_coords[index];
}

void CoalShape::RemoveCoord (unsigned int index)
{
    printf ("FIXME: CoalShape::RemoveCoord() not implemented!!\n");

    // FIXME: Hmmm, is this right?...
    //if (index >= 0 && index < m_coords.size ())
    //    m_coords.erase (m_coords[index]);
}

void CoalShape::ChangeCoord (unsigned int index, const CoalCoord &newCoord)
{
    if (index >= 0 && index < m_coords.size ())
        m_coords[index] = newCoord;
}

void CoalShape::MakeCircle (const CoalCoord &center, CoalCoordType radius)
{
    ClearShape ();
    m_type = COAL_SHAPE_CIRCLE;
    m_coords[0] = center;
    m_length = radius;
}

void CoalShape::MakeSquare (const CoalCoord &origin, CoalCoordType edgelength)
{
    ClearShape ();
    m_type = COAL_SHAPE_SQUARE;
    m_coords[0] = origin;
    m_coords[1].SetX (origin.GetX () + edgelength);
    m_coords[1].SetY (origin.GetY () + edgelength);
}

void CoalShape::MakeRect (const CoalRect &rect)
{
    ClearShape ();
    m_type = COAL_SHAPE_RECT;
    m_coords.push_back (rect.GetTopLeft());
    m_coords.push_back (rect.GetBottomRight());
}

void CoalShape::MakeRect (const CoalCoord &origin,
                           CoalCoordType width, CoalCoordType height)
{
    ClearShape ();
    m_type = COAL_SHAPE_RECT;
    m_coords.push_back (origin);

    CoalCoord bottomright (origin.GetX () + width - 1,
                           origin.GetY () + height - 1);
    m_coords.push_back (bottomright);
}


// Get the bounding rectangle that contains the entire region
void CoalShape::GetBoundingRect (CoalRect &rect) const
{
    // Set starting point to the first coord in the shape, rather
    // than the default value of (0, 0, 0)
    CoalCoord coord = GetCoord (0);
    rect.SetRect (coord, coord);

    if (m_verbose)
    {       
        printf ("---GetBoundingRect()\n--- ");
        CoalDebug::Dump (coord);
    }

    int count = GetCoordCount ();
    for (int i = 1; i < count; i++)
    {
        coord = GetCoord (i);
        if (m_verbose)
        {
            printf ("--- ");
            CoalDebug::Dump (coord);
        }

        rect.Intersect (coord);
    }
}

// Is coord equal to one of the vertices?
bool CoalShape::IsPoint (const CoalCoord &coord) const
{
    int count = GetCoordCount ();
    for (int i = 0; i < count; i++)
    {
        // FIXME: g++ complains about the const here...hmmmm...
        if (m_coords[i] == coord)
            return true;
    }

    return false;
}

bool CoalShape::IsPointIn (const CoalCoord &coord) const
{
    // This function is the high point of the CoalShape class.
    // It'll take some clever algorithms to implement it.  Hopefully
    // the CoalShapeType enum will help optimize things.

    bool success;

    switch (m_type)
    {
    case COAL_SHAPE_CIRCLE:
        success = IsPointInCircle (coord);
        break;

    case COAL_SHAPE_SQUARE:  // May someday optimize for squares...?
    case COAL_SHAPE_RECT:
        success = IsPointInRect (coord);
        break;

    case COAL_SHAPE_POLYGON:
        success = IsPointInPolygon (coord);
        break;

    default:
        success = false;
        break;
    }

    return success;
}

bool CoalShape::IsPointInCircle (const CoalCoord &coord) const
{
    CoalCoord center = GetCoord (0);
    if (!center.IsValid ())
        return false;

    return (center.GetDistanceTo (coord) <= m_length);
}

bool CoalShape::IsPointInRect (const CoalCoord &coord) const
{
    CoalRect rect (m_coords[0], m_coords[1]);

    return (coord.GetY () >= rect.GetTop () &&
            coord.GetX () >= rect.GetLeft () &&
            coord.GetY () < rect.GetBottom () &&
            coord.GetX () < rect.GetRight ());
}

// Polygon hit-testing algorithm modified from:
// 
//  http://www.ecse.rpi.edu/Homepages/wrf/geom/pnpoly.html
//  
bool CoalShape::IsPointInPolygon (const CoalCoord &coord) const
{
    //int pnpoly(int npol, float *xp, float *yp, float x, float y)

    int point_count = m_coords.size ();

    int i, j;
    bool hit = false;
    bool vertically_between;

    for (i = 0, j = point_count - 1; i < point_count; j = i++)
    {
        vertically_between =
            (((m_coords[i].GetY() <= coord.GetY()) &&
              (coord.GetY() < m_coords[j].GetY())) ||
             ((m_coords[j].GetY() <= coord.GetY()) &&
              (coord.GetY() < m_coords[i].GetY())));

        if (vertically_between &&
            (coord.GetX() < (m_coords[j].GetX() - m_coords[i].GetX()) *
             (coord.GetY() - m_coords[i].GetY()) /
             (m_coords[j].GetY() - m_coords[i].GetY()) + m_coords[i].GetX()))
        {
            hit = !hit;
        }
    }

    if (0)
        printf ("Found polygon hit at (%.2f, %.2f)\n",
                coord.GetX(), coord.GetY());

    return hit;
}

// ********** CoalRect implementation **********

void CoalRect::SetRect (CoalCoordType top, CoalCoordType left,
                        CoalCoordType bottom, CoalCoordType right)
{
    m_topleft.SetCoord (left, top, 0);
    m_bottomright.SetCoord (right, bottom, 0);
}

void CoalRect::SetRect (const CoalCoord &topleft,
                        const CoalCoord &bottomright)
{
    m_topleft = topleft;
    m_bottomright = bottomright;
}

void CoalRect::Intersect (const CoalRect &rect)
{
    if (rect.GetLeft() < GetLeft())
       SetLeft (rect.GetLeft());
    if (rect.GetTop() < GetTop())
       SetTop (rect.GetTop());

    if (rect.GetRight() > GetRight())
       SetRight (rect.GetRight());
    if (rect.GetBottom() > GetBottom())
       SetBottom (rect.GetBottom());
}

void CoalRect::Intersect (const CoalCoord &coord)
{
    if (coord.GetX() < GetLeft())
        SetLeft (coord.GetX());
    if (coord.GetY() < GetTop())
        SetTop (coord.GetY());

    if (coord.GetX() > GetRight())
        SetRight (coord.GetX());
    if (coord.GetY() > GetBottom())
        SetBottom (coord.GetY());
}

// Is this point inside the rectangle?
bool CoalRect::IsPointIn (const CoalCoord &coord) const
{
    return (coord.GetX() >= GetLeft() && coord.GetX() <= GetRight() && 
            coord.GetY() >= GetTop() && coord.GetY() <= GetBottom());
}

// Shrink both directions by the same amount
void CoalRect::Shrink (CoalCoordType shrinkby)
{
    Shrink (shrinkby, shrinkby);
}

// Shrink towards the middle
void CoalRect::Shrink (CoalCoordType width, CoalCoordType height)
{
    m_topleft += CoalCoord (width / 2.0, height / 2.0);
    m_bottomright -= CoalCoord (width / 2.0, height / 2.0);
}

// Shrink both directions by the same amount
void CoalRect::Expand (CoalCoordType shrinkby)
{
    Shrink (shrinkby, shrinkby);
}

// Shrink towards the middle
void CoalRect::Expand (CoalCoordType width, CoalCoordType height)
{
    m_topleft -= CoalCoord (width / 2.0, height / 2.0);
    m_bottomright += CoalCoord (width / 2.0, height / 2.0);
}
