// -*-C++-*-
// shape.h
//
//  The WorldForge Project
//  Copyright (C) 2000  The WorldForge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about WorldForge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//

#ifndef __SHAPE_H
#define __SHAPE_H

#include "coord.h"

class CoalRect;

typedef enum
{
    COAL_SHAPE_CIRCLE,
    COAL_SHAPE_TRIANGLE,
    COAL_SHAPE_SQUARE,
    COAL_SHAPE_RECT,
    COAL_SHAPE_POLYGON
} CoalShapeType;

class CoalShape
{
public:
    CoalShape () { ClearShape (); SetVerbose (false); }
    CoalShape (CoalShapeType shape) { m_type = shape; }
    virtual ~CoalShape ();

    // Accessors
    void SetType (CoalShapeType rtype) { m_type = rtype; }
    CoalShapeType GetType () const { return m_type; }
    void SetLength (CoalCoordType length) { m_length = length; }
    CoalCoordType GetLength () const { return m_length; }

    void ClearShape ();

    unsigned int GetCoordCount () const { return m_coords.size (); }
    void AddCoord (const CoalCoord &coord) { m_coords.push_back (coord); }

    CoalCoord GetCoord (unsigned int index) const;
    void RemoveCoord (unsigned int index);
    void ChangeCoord (unsigned int index, const CoalCoord &newCoord);

    // May later add a MakePolygon ()
    //void MakePolygon (const CoalCoord &coord1, ...);

    // Convenience functions
    void MakeCircle (const CoalCoord &center, CoalCoordType radius);
    void MakeSquare (const CoalCoord &topleft, CoalCoordType edgelength);
    void MakeRect (const CoalRect &rect);
    void MakeRect (const CoalCoord &origin,
                   CoalCoordType width, CoalCoordType height);

    // Hit testing
    bool IsPoint (const CoalCoord &coord) const;
    bool IsPointIn (const CoalCoord &coord) const;
    void GetBoundingRect (CoalRect &rect)  const;

    void SetVerbose (bool verbose) { m_verbose = verbose; }

private:
    bool IsPointInCircle (const CoalCoord &coord) const;
    bool IsPointInRect (const CoalCoord &coord) const;
    bool IsPointInPolygon (const CoalCoord &coord) const;

private:
    bool m_verbose;

    CoalShapeType m_type;
    CoalCoordArray m_coords;

    CoalCoordType m_length;  // Used with COAL_REGION_CIRCLE & SQUARE

    // Possible other fields for expressing regular polygons...

    //CoalAngle m_angle;  // Angle of first line segment in a regular polygon,
    //                    // like a pentagon or octagon
    //int m_segments;
};

// Put this here for now.  Not sure if CoalShape should totally
// replace this class, or if it's worthwhile to keep the rectangle
// specialization around...
class CoalRect
{
public:
    CoalRect () { ClearRect (); }
    CoalRect (const CoalCoord &topleft, const CoalCoord &bottomright)
        { SetRect (topleft, bottomright); }
    CoalRect (CoalCoordType top, CoalCoordType left,
              CoalCoordType bottom, CoalCoordType right)
        { SetRect (top, left, bottom, right); }

    virtual ~CoalRect () {}
    CoalRect& operator= (const CoalRect &copy)
        { m_topleft = copy.GetTopLeft ();
          m_bottomright = copy.GetBottomRight ();
          return *this; }

    void Intersect (const CoalRect &rect);
    void Intersect (const CoalCoord &coord);

    bool IsPointIn (const CoalCoord &coord) const;
    CoalCoordType Width () const { return GetRight () - GetLeft (); }
    CoalCoordType Height () const { return GetBottom () - GetTop (); }

    // Info functions
    CoalCoordType GetTop () const { return m_topleft.GetY (); }
    CoalCoordType GetLeft () const { return m_topleft.GetX (); }
    CoalCoordType GetBottom () const { return m_bottomright.GetY (); }
    CoalCoordType GetRight () const { return m_bottomright.GetX (); }

    CoalCoord GetTopLeft () const
        { return m_topleft; }
    CoalCoord GetBottomRight () const
        { return m_bottomright; }

    void SetTop (CoalCoordType top) { m_topleft.SetY (top); }
    void SetLeft (CoalCoordType left) { m_topleft.SetX (left); }
    void SetBottom (CoalCoordType bottom) { m_bottomright.SetY (bottom); }
    void SetRight (CoalCoordType right) { m_bottomright.SetX (right); }

    void SetTopLeft (const CoalCoord &topleft)
        { m_topleft = topleft; }
    void SetBottomRight (const CoalCoord &bottomright)
        { m_bottomright = bottomright; }

    void SetRect (CoalCoordType top, CoalCoordType left,
                  CoalCoordType bottom, CoalCoordType right);
    void SetRect (const CoalCoord &topleft,
                  const CoalCoord &bottomright);

    void ClearRect () { SetRect (0, 0, 0, 0); }
    void Shrink (CoalCoordType width, CoalCoordType height);
    void Shrink (CoalCoordType width);
    void Expand (CoalCoordType width, CoalCoordType height);
    void Expand (CoalCoordType width);

private:
    CoalCoord m_topleft;
    CoalCoord m_bottomright;
};

typedef vector<CoalShape*> CoalShapeArray;
typedef vector<CoalRect*> CoalRectArray;

#endif  // __SHAPE_H
